import tensorflow as _tf
import numpy as _np


def mean_absolute_scaled_error(y_true, y_pred):
    """
    Implement MASE (assuming no seasonality of data).
    """
    mae = _tf.reduce_mean(_tf.abs(y_true - y_pred))

    # Find MAE of naive forecast (no seasonality)
    mae_naive_no_season = _tf.reduce_mean(_tf.abs(y_true[1:] - y_true[:-1]))

    return mae / mae_naive_no_season


def evaluate_preds(y_true, y_pred) -> dict:
    y_true = _tf.cast(y_true, dtype=_tf.float32)
    y_pred = _tf.cast(y_pred, dtype=_tf.float32)

    mse = _tf.keras.metrics.mean_squared_error(y_true, y_pred)
    rmse = _np.sqrt(mse)

    if not isinstance(rmse, _np.floating):
        rmse = rmse.numpy()

    return {
        'mae': _tf.keras.metrics.mean_absolute_error(y_true, y_pred).numpy(),
        'mse': mse.numpy(),
        'rmse': rmse,
        'mape': _tf.keras.metrics.mean_absolute_percentage_error(y_true, y_pred).numpy(),
        'mase': mean_absolute_scaled_error(y_true, y_pred).numpy()
    }



import matplotlib.pyplot as _plt
import seaborn as _sns
import itertools as _itertools
import numpy as _np
import tensorflow as _tf


def timeseries(x: list[_np.ndarray],
               y: list[_np.ndarray],
               labels: list[str] = None,
               title: str = None,
               x_label: str = None,
               y_label: str = None,
               figsize: tuple = (8, 6),
               label_color='black') -> None:
    """
    Plots time series data, where `x` is a single array containing the date/time and `y` can
    be one or more arrays of labels and predictions.

    In case `y` has a larger horizon then one, for example to predict 7 days, then the data will be
    reduced using the mean operation.

    The use this method you can call it in the following ways:

    timeseries(
        x=[
            X_history,
            X_pred
        ],
        y=[
            y_history,
            y_pred
        ],
        labels=['Label', 'Pred'],
        start=300,
        x_label='Time', y_label
        ='Price')

    Note that filling in the values for X and y might be different in certain circumstances, most often this has
    to do with the shape, make sure they match.

    Where day is set to the day in the range of the horizon, in this case the third day.

    Args:
        x: A list containing one ore more arrays of dates/times
        y: A list containing one ore more arrays of labels and predictions
        labels: An array of labels where each label corresponds to an entry in the list of y.
        title: An optional title for the graph.
        x_label: An optional x-axis label for the graph.
        y_label: An optional y-axis label for the graph.
        figsize: An optional figure size. Default is (8, 6)
        label_color: An optional color for the labels of the axis, axis values and title.
    """
    fig, ax = _plt.subplots(figsize=figsize)

    color_palette = _itertools.cycle(_sns.color_palette(palette="tab10", n_colors=len(x)))
    for idx, _y in enumerate(y):
        _x = x[idx]

        # If the horizon of the data contains more then one day then reduce the
        if hasattr(_y, 'ndim') and _y.ndim > 1:
            _y = _tf.reduce_mean(_y, axis=1)
            if labels is not None and len(labels) > idx:
                labels[idx] = f"{labels[idx]} (mean)"
        else:
            _y = y[idx]

        _label = 'None'
        if labels is not None and len(labels) > idx:
            _label = labels[idx - 1]



        _sns.lineplot(
            x=_x,
            y=_y,
            color=next(color_palette),
            # s=7,
            ax=ax,
            label=_label)

    if x_label:
        _plt.xlabel(x_label)
    if y_label:
        _plt.ylabel(y_label)
    if title:
        _plt.title(title)

    fig.patch.set_alpha(0.0)  # Transparant background

    ax.tick_params(axis='x', labelrotation=90)

    ax.xaxis.label.set_color(label_color)  # Set color of x-axis label
    ax.tick_params(axis='x', colors=label_color)  # Set color of x-axis ticks.

    ax.yaxis.label.set_color(label_color)  # Set color of y-axis label
    ax.tick_params(axis='y', colors=label_color)  # Set color of y-axis ticks.
    ax.title.set_color(label_color)  # Set color of title

    if labels:
        _plt.legend()

    _plt.show()



import tensorflow as _tf


def mean_absolute_scaled_error(y_true, y_pred, seasonality: int = None):
    """
    Implement MASE (assuming no seasonality of data).
    """
    mae = _tf.reduce_mean(_tf.abs(y_true - y_pred))

    #  Mean Absolute Deviation (MAD) of the historical data
    if seasonality is not None:
        mad = _tf.reduce_mean(_tf.abs(y_true[seasonality:] - y_true[:-seasonality]))
    else:
        mad = _tf.reduce_mean(_tf.abs(y_true[1:] - y_true[:-1]))

    return mae / mad



import random as _random
import os as _os
import logging as _logging

import tensorflow as _tf
import matplotlib.pyplot as _plt
import so_ml_tools as _soml


def load_image_as_tensor(filename: str, img_shape=(224, 224), scale=True) -> _tf.Tensor:
    """
    Reads in an image from filename, turns it into a tensor and reshapes into
    specified shape (img_shape, img_shape, channels)

    Args:
        filename: path to target image
        img_shape: A `tuple` with the dimensions of the image.
        scale: A `bool` indicating if scaling needs to be performed (0-255 -> 0-1)

    Returns:
        A 'tf.Tensor' of shape (img_shape, 3) with of dtype 'tf.int32'.
    """

    img = _tf.io.read_file(filename=filename)
    return image_as_tensor(image=img, img_shape=img_shape, scale=scale)


def image_as_tensor(image, img_shape: tuple = (224, 224), scale: bool = True) -> _tf.Tensor:
    """
    Reads in an image from filename, turns it into a tensor and reshapes into
    specified shape (img_shape, img_shape, channels)
            
    Args:
        image: A `Tensor` of type `string`. 0-D. The encoded image bytes.
        img_shape: A `tuple` with the dimensions of the image.
        scale: A `bool` indicating if scaling needs to be performed (0-255 -> 0-1)

    Returns:
        A 'tf.Tensor' of shape (img_shape, 3) with of dtype 'tf.int32'.
    """
    # Decode image into tensor
    img = _tf.io.decode_image(contents=image, channels=3)

    # # Resize the image (height / width)
    img = _tf.image.convert_image_dtype(img, _tf.float32)
    img = _tf.image.resize(images=img, size=[img_shape[0], img_shape[1]])

    if not scale:
        return _tf.cast(x=img * 255, dtype=_tf.int32)
    else:
        return img


def is_image_float32_and_not_normalized(x) -> bool:
    """
    Convenience method to check if an image in `tf.Tensor` format is of type `float32` but not normalized (values between
    0..255 instead of 0..1)

    Args:
        x: A `tf.Tensor`

    Returns:
        A `bool` indicating if the tensor is of type `float32` but not normalized.
    """
    return x.dtype == _tf.float32 and _tf.math.reduce_max(x).numpy() > 1.0


def show_images_from_nparray_or_tensor(x, y, class_names: list[str] = None, indices=None, shape: tuple = (4, 6),
                                       cmap: str = 'gray') -> None:
    """
    Shows images stored in a tensor / numpy array. The array should be a vector of images.
    
    Args:
        x: a `tf.Tensor` containg the images
        y: a 'tf.Tensor' containing the labels associated with `x`
        class_names: a `list` of class names 
        indices: a `list` of indices for the images to display, `None` to pick random images
        shape: A `tuple` specifying the number of images to display (width, height)
        cmap: A 'str' with the color map to use for displaying the images.

    Returns:

    """
    y = _soml.util.label.to_prediction(y_prob=y)

    if is_image_float32_and_not_normalized(x):
        x = _tf.cast(x=x, dtype=_tf.uint8)

    if indices:
        assert shape[0] * shape[1] <= len(
            indices), f"Size of shape ({shape[0]}, {shape[1]}), with a total of " \
                      f"{shape[0] * shape[1]} images, is larger then number of indices supplied ({len(indices)})."
        for i in indices:
            if i > len(x):
                assert False, f"Values of indices point to an index ({i}) which is out of bounds of X (length: {len(x)})"

    fig = _plt.figure(figsize=(shape[1] * 3, shape[0] * 3))
    fig.patch.set_facecolor('gray')
    for i in range(shape[0] * shape[1]):
        ax = _plt.subplot(shape[0], shape[1], i + 1)
        ax.axis('off')

        if indices is None:
            rand_index = _random.choice(range(len(x)))
        else:
            rand_index = indices[i]

        _plt.imshow(x[rand_index], cmap=cmap)

        if _soml.util.label.is_multiclass_classification(y):
            class_index = _soml.util.label.probability_to_class(y)
        else:
            # Integer encoded labels
            class_index = y[rand_index]

        if class_names is None:
            _plt.title(class_index, color='white')
        else:
            _plt.title(f"{class_names[class_index]}, {x[rand_index].shape}", color='white')
    _plt.show()


def show_single_image_from_nparray_or_tensor(image, title="", figsize=(10, 8), cmap='gray') -> None:
    """
    Shows images stored in a tensor / numpy array. The array should be a vector of images.

    Args:
        image: the image to plot
        title: the title to display above the image
        figsize: Size of output figure (default=(10, 8)).
        cmap: is the collor map to use, use "gray" for gray scale images, use None for default.

    Returns:
        None
    """
    fig = _plt.figure(figsize=figsize)
    fig.patch.set_facecolor('gray')
    _plt.imshow(image, cmap=cmap)
    _plt.axis('off')
    _plt.title(f"{title} {image.shape}", color='white')
    _plt.show()


def show_random_image_from_disk(target_dir, target_class, shape=(4, 6), cmap='gray') -> None:
    """
    Shows a random image from the file system.

    Args:
        target_dir: The target directory, for example /food101/train
        target_class: The target class name, for example /steak
        shape: is the number of images in a grid to display
        cmap: is the color map to use, use "gray" for gray scale images, use None for default.

    Returns:
        None
    """
    try:
        target_folder = _os.path.join(target_dir, target_class)

        # Get a random image path
        fig = _plt.figure(figsize=(shape[1] * 3, shape[0] * 3))
        fig.patch.set_facecolor('gray')
        for i in range(0, shape[0] * shape[1]):
            ax = _plt.subplot(shape[0], shape[1], i + 1)
            ax.axis('off')
            image_filename = _random.sample(_os.listdir(target_folder), 1)
            image_fullpath = _os.path.join(target_folder, image_filename[0])
            img = _plt.imread(image_fullpath)
            _plt.imshow(img, cmap=cmap)
            _plt.title(f"{image_filename[0]}: {img.shape}", color='white')

        _plt.show()
    except Exception as e:
        _logging.error(e)


def show_images_from_dataset(dataset: _tf.data.Dataset, shape=(4, 8)) -> None:
    """
    Displays a number of images based on the shape size, the shape size must be the same
    as the batch size of the dataset. Default shape size is (4, 8) so the batch size
    should be 32.

    Args:
        dataset: the dataset containing the images
        shape: the shape size

    Returns:
        None
    """
    assert isinstance(dataset, _tf.data.Dataset), f"The dataset supplied is not a tensorflow.data.Dataset."

    # Retrieve first batch, depending on the initalization of the dataset the batch size is default 32
    # so when performing a take of (1) we retreive the first batch
    batches = dataset.take(1)

    # Use an iterator to get the first batch of images and labels
    batch_iter = iter(batches)
    x, y_true = batch_iter.next()

    assert shape[0] * shape[1] == len(x), f"Size of shape ({shape[0]}, {shape[1]}), with a total of " \
                                          f"{shape[0] * shape[1]} images, is not equal to the batch size of the dataset ({len(x)})."

    class_names = _soml.tf.dataset.get_class_names(dataset=dataset)
    show_images_from_nparray_or_tensor(x=x, y=y_true, class_names=class_names, shape=shape)
    _plt.show()



import numpy as _np


# Create function to view NumPy arrays as windows
def make_windows(x, window_size, horizon):
    """
    Turns 1 1D array into a 2D array of sequential labelled windows of window_size with horizon size labels.
    """
    # 1. Create a window of specific window_size (add the horizon on the end for labelling later)
    window_step = _np.expand_dims(_np.arange(window_size+horizon), axis=0)

    # 2. Create a 2D array for multiple window step (minus 1 to account for 0 indexing)
    window_indexes = window_step + _np.expand_dims(_np.arange(len(x)-(window_size + horizon - 1)), axis=0).T # Create 2D array of windows of window size window_size.

    # 3. Index on the target arraay (a time series) with 2D array of multiple widnow steps
    windowed_array = x[window_indexes]

    #4. get the labelled widnows
    windows, labels = windowed_array[:, :-horizon], windowed_array[:, -horizon:]
    return windows, labels



import pandas as _pd
import numpy as _np

from sklearn.preprocessing import OneHotEncoder


########################################################################################################################
# General
########################################################################################################################

def simple_undersample(dataframe: _pd.DataFrame, label_column, value_to_undersample, undersample_to):
    """
    Undersample a class (label_column) with a value (value_to_undersample) to the number of entries (undersample_to)

    For example:

    undersample(dataframe: training_data, column_name='stroke', value_to_undersample=0, undersample_to=2000)

    Args:
        dataframe: the pd.DataFrame
        label_column: the column to match on
        value_to_undersample: the value within the `column` to match on for under-sampling.
        undersample_to: number of entries that need to exist after the under-sampling.

    Returns:
        A new dataframe with the data.
    """
    print('A better solution for under-sampling / over-sampling is to use the soml.imblearn.smote package.')
    unique = dataframe[label_column].unique()

    sections = {}
    for v in unique:
        sections[v] = dataframe[dataframe[label_column] == v]

    sections[value_to_undersample] = sections[value_to_undersample].sample(undersample_to)
    return _pd.concat(sections.values(), axis=0)


def simple_oversample(dataframe: _pd.DataFrame, label_column, value_to_oversample, oversample_to):
    """
    oversample a class (label_column) with a value (value_to_oversample) to the number of entries (oversample_to)

    For example:

    oversample(dataframe: training_data, column_name='stroke', value_to_oversample=1, oversample_to=2000)

    Args:
        dataframe: the pd.DataFrame
        label_column: the column to match on
        value_to_oversample: the value within the `column` to match on for over sampling.
        oversample_to: number of entries that need to exist after the over sampling.

    Returns:
        A new dataframe with the data.
    """
    print('A better solution for under-sampling / over-sampling is to use the soml.imblearn.smote package.')
    unique = dataframe[label_column].unique()

    sections = {}
    for v in unique:
        sections[v] = dataframe[dataframe[label_column] == v]

    sections[value_to_oversample] = sections[value_to_oversample].sample(oversample_to, replace=True)
    return _pd.concat(sections.values(), axis=0)


def one_hot_encode_column(*columns: list[any]) -> (OneHotEncoder, list[any]):
    """
    One-Hot encodes all given columns using all values of all the provided columns.

    Example usage:

    encoder, train_labels_one_hot, val_labels_one_hot, test_labels_one_hot = \
    one_hot_encode_column(
        train_df['target'],
        val_df['target'],
        test_df['target']
    )

    If you get the ValueError:

    all the input arrays must have same number of dimensions, but the array at index 0 has 1 dimension(s) and the
    array at index 1 has 2 dimension(s)

    Then that means one of the lists passed in has ha different shape then the others.

    :param columns: one or more columns
    :return: first value is the encoder, all other values are in same order as the columns but encoded
    """
    encoder = OneHotEncoder(sparse_output=False)

    # Preprocess the columns
    preprocessed_columns = []
    for column in columns:
        if isinstance(column, _pd.DataFrame) or isinstance(column, _pd.Series):
            column = column.to_numpy()

        # Make sure the column is two-dimensional
        if column.ndim == 1:
            column = column.reshape(-1, 1)

        preprocessed_columns.append(column)

    # Concatenate all columns and fit the encoder
    encoder.fit(_np.concatenate(preprocessed_columns))

    return_values = []
    # Transform each list now individually
    for preprocessed_column in preprocessed_columns:
        return_values.append(encoder.transform(preprocessed_column))

    return_values.insert(0, encoder)
    return return_values


def column_as_dataframe(dataframe: _pd.DataFrame, column_name: str, drop_after: False) -> _pd.DataFrame:
    column = dataframe[[column_name]]

    if drop_after:
        drop_columns(dataframe=dataframe, column_names=[column_name], inplace=True)

    return column


def convert_column_to_type(dataframe: _pd.DataFrame, columns: list[str], dtype=_np.float64 | str,
                           inplace=True) -> _pd.DataFrame | None:
    """
    Converts the dtype of the given column to the specified dtype.

    :param dataframe: the pd.DataFrame
    :param columns: A `list` of column names
    :param dtype: the newly to assign dtype.
    :param inplace: return a new instance of the DataFrame or adjust the given DataFrame
    :return: see inplace
    """
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    for column in columns:
        work_df[column] = work_df[column].astype(dtype)

    if inplace:
        return None

    return work_df


def delete_rows_where_columns_have_null_value(dataframe: _pd.DataFrame, column_names: list[str] = None, inplace=True) \
        -> _pd.DataFrame | None:
    """
    Deletes all rows containing a null values inside the given column .

    :param dataframe: the pd.DatFrame
    :param column_names: the names of the columns to check, if None all columns will be checked.
    :param inplace: return a new instance of the DataFrame (False) or adjust the given DataFrame
    :return: see inplace
    """
    if column_names is None:
        column_names = list(dataframe.columns)

    if not type(column_names) is list and column_names is not None:
        column_names = [column_names]

    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    for c in column_names:
        if c in dataframe:
            work_df.drop(work_df[dataframe[c].isnull()].index, inplace=True)
        else:
            print(f"delete_null_rows: Column '{c}' does not exist in dataframe.")

    if inplace:
        return None

    return work_df


def delete_rows_where_value_equal_to(dataframe: _pd.DataFrame, column_name: str, value: object, inplace=True) \
        -> _pd.DataFrame | None:
    """
    Deletes all rows where the specified column has the specified value.

    :param dataframe: the pd.DatFrame
    :param column_name: the names of the column to check
    :param value: the value to compare with
    :param inplace: return a new instance of the DataFrame (False) or adjust the given DataFrame
    :return: see inplace
    """
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    if column_name in dataframe:
        work_df.drop(dataframe[dataframe[column_name] == value].index, inplace=True)
    else:
        print(f"delete_rows_where_value_equal_to: Column '{column_name}' does not exist in dataframe.")

    if inplace:
        return None

    return work_df


def delete_rows_where_value_smaller_then(dataframe: _pd.DataFrame, column_name: str, value: object, inplace=True) \
        -> _pd.DataFrame | None:
    """
    Deletes all rows where the specified column has the specified value.

    :param dataframe: the pd.DatFrame
    :param column_name: the names of the column to check
    :param value: the value to compare with
    :param inplace: return a new instance of the DataFrame (False) or adjust the given DataFrame
    :return: see inplace
    """
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    if column_name in dataframe:
        work_df.drop(dataframe[dataframe[column_name] < value].index, inplace=True)
    else:
        print(f"delete_rows_where_value_equal_to: Column '{column_name}' does not exist in dataframe.")

    if inplace:
        return None

    return work_df


def delete_rows_where_value_larger_then(dataframe: _pd.DataFrame, column_name: str, value: object, inplace=True) \
        -> _pd.DataFrame | None:
    """
    Deletes all rows where the specified column has the specified value.

    :param dataframe: the pd.DatFrame
    :param column_name: the names of the column to check
    :param value: the value to compare with
    :param inplace: return a new instance of the DataFrame (False) or adjust the given DataFrame
    :return: see inplace
    """
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    if column_name in dataframe:
        work_df.drop(dataframe[dataframe[column_name] > value].index, inplace=True)
    else:
        print(f"delete_rows_where_value_equal_to: Column '{column_name}' does not exist in dataframe.")

    if inplace:
        return None

    return work_df


def fill_nan_with_value(dataframe: _pd.DataFrame, column_names: list[str], value, inplace=True, add_indicator=False) \
        -> _pd.DataFrame | None:
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    if type(column_names) is not list and column_names is not None:
        column_names = [column_names]

    for c in column_names:
        if c in dataframe:
            if work_df[c].isna().any():
                if add_indicator:
                    work_df[c + '_nan'] = work_df[c].isna().astype(int)
                work_df[c].fillna(value=value, inplace=True)
        else:
            print(f"fill_nan_with_value: Column '{c}' does not exist in dataframe.")

    if inplace:
        return None

    return work_df


def fill_nan_with_previous_value(dataframe: _pd.DataFrame, column_names: list[str], inplace=True, add_indicator=False) \
        -> _pd.DataFrame | None:
    """
        Fill in missing values based on the previous value.

        Args:
            dataframe: the pd.DatFrame
            column_names: list of column names with `NaN` values.
            inplace: update the given dataframe or return a new dataframe.
            add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    if type(column_names) is not list and column_names is not None:
        column_names = [column_names]

    for c in column_names:
        if c in dataframe:
            if add_indicator:
                work_df[c + '_nan'] = work_df[c].isna().astype(int)
            work_df[c].fillna(method='ffill', inplace=True)
        else:
            print(f"fill_nan_with_previous_value: Column '{c}' does not exist in dataframe.")

    if inplace:
        return None

    return work_df


def fill_nan_with_next_value(dataframe: _pd.DataFrame, column_names: list[str], inplace=True, add_indicator=False) \
        -> _pd.DataFrame | None:
    """
        Fill in missing values based on the next value.

        Args:
            dataframe: the pd.DatFrame
            column_names: list of column names with `NaN` values.
            inplace: update the given dataframe or return a new dataframe.
            add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    if type(column_names) is not list and column_names is not None:
        column_names = [column_names]

    for c in column_names:
        if c in dataframe:
            if add_indicator:
                work_df[c + '_nan'] = work_df[c].isna().astype(int)
            work_df[c].fillna(method='bfill', inplace=True)
        else:
            print(f"fill_nan_with_next_value: Column '{c}' does not exist in dataframe.")

    if inplace:
        return None

    return work_df


def fill_nan_with_global_mode(dataframe: _pd.DataFrame, column_names: list[str], inplace=True,
                              add_indicator=False) -> _pd.DataFrame | None:
    """
        Fill in missing values based on the mode (most often) value which is calculated on all non `NaN` values
        in the same column.

        Args:
            dataframe: the pd.DatFrame
            column_names: list of column names with `NaN` values.
            inplace: update the given dataframe or return a new dataframe.
            add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_global_func(dataframe=dataframe, column_names=column_names, agg_func='mode', inplace=inplace,
                                     add_indicator=add_indicator)


def fill_nan_with_global_kurt(dataframe: _pd.DataFrame, column_names: list[str], inplace=True,
                              add_indicator=False) -> _pd.DataFrame | None:
    """
        Fill in missing values based on kurt value which is calculated on all non `NaN` values in the same
        column.

        Args:
            dataframe: the pd.DatFrame
            column_names: list of column names with `NaN` values.
            inplace: update the given dataframe or return a new dataframe.
            add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_global_func(dataframe=dataframe, column_names=column_names, agg_func='kurt', inplace=inplace,
                                     add_indicator=add_indicator)


def fill_nan_with_global_skew(dataframe: _pd.DataFrame, column_names: list[str], inplace=True,
                              add_indicator=False) -> _pd.DataFrame | None:
    """
        Fill in missing values based on skew value which is calculated on all non `NaN` values in the same
        column.

        Args:
            dataframe: the pd.DatFrame
            column_names: list of column names with `NaN` values.
            inplace: update the given dataframe or return a new dataframe.
            add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_global_func(dataframe=dataframe, column_names=column_names, agg_func='kurt', inplace=inplace,
                                     add_indicator=add_indicator)


def fill_nan_with_global_median(dataframe: _pd.DataFrame, column_names: list[str], inplace=True,
                                add_indicator=False) -> _pd.DataFrame | None:
    """
        Fill in missing values based on median value which is calculated on all non `NaN` values in the same
        column.

        Args:
            dataframe: the pd.DatFrame
            column_names: list of column names with `NaN` values.
            inplace: update the given dataframe or return a new dataframe.
            add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_global_func(dataframe=dataframe, column_names=column_names, agg_func='median', inplace=inplace,
                                     add_indicator=add_indicator)


def fill_nan_with_global_mean(dataframe: _pd.DataFrame, column_names: list[str], inplace=True,
                              add_indicator=False) -> _pd.DataFrame | None:
    """
        Fill in missing values based on mean value which is calculated on all non `NaN` values in the same
        column.

        Args:
            dataframe: the pd.DatFrame
            column_names: list of column names with `NaN` values.
            inplace: update the given dataframe or return a new dataframe.
            add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_global_func(dataframe=dataframe, column_names=column_names, agg_func='mean', inplace=inplace,
                                     add_indicator=add_indicator)


def fill_nan_with_global_max(dataframe: _pd.DataFrame, column_names: list[str], inplace=True, add_indicator=False) \
        -> _pd.DataFrame | None:
    """
        Fill in missing values based on max value which is calculated on all non `NaN` values in the same
        column.

        Args:
            dataframe: the pd.DatFrame
            column_names: list of column names with `NaN` values.
            inplace: update the given dataframe or return a new dataframe.
            add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_global_func(dataframe=dataframe, column_names=column_names, agg_func='max', inplace=inplace,
                                     add_indicator=add_indicator)


def fill_nan_with_global_min(dataframe: _pd.DataFrame, column_names: list[str], inplace=True,
                             add_indicator=False) -> _pd.DataFrame | None:
    """
        Fill in missing values based on min value which is calculated on all non `NaN` values in the same
        column.

        Args:
            dataframe: the pd.DatFrame
            column_names: list of column names with `NaN` values.
            inplace: update the given dataframe or return a new dataframe.
            add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_global_func(dataframe=dataframe, column_names=column_names, agg_func='min', inplace=inplace,
                                     add_indicator=add_indicator)


def fill_nan_with_global_func(dataframe: _pd.DataFrame, column_names: list[str], agg_func: str, inplace=True,
                              add_indicator=False) -> _pd.DataFrame | None:
    """
        Fill in missing values based on a func value which is calculated on all non `NaN` values in the same
        column.

        Args:
            dataframe: the pd.DatFrame
            column_names: list of column names with `NaN` values.
            agg_func: type of function to use: 'mean', 'min', 'max', 'median', 'skew' or 'kurt'
            inplace: update the given dataframe or return a new dataframe.
            add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    for c in column_names:
        if c in dataframe:
            if add_indicator:
                work_df[c + '_nan'] = work_df[c].isna().astype(int)

            c_value = work_df[c].apply(agg_func)
            work_df[c].fillna(value=c_value, inplace=True)
        else:
            print(f"fill_nan_with_mean: Column '{c}' does not exist in dataframe.")

    if inplace:
        return None

    return work_df


def ffill_nan_with_groupby_mode(dataframe: _pd.DataFrame, column_name: str, group_by_column_name: str,
                                  remainder_agg_func='mode', inplace=True, add_indicator=False) -> _pd.DataFrame | None:
    """
    Fill in missing values based on the mode value (most often) which is calculated on all data having a similar value
    in another column. For example, fill in BMI values based on other BMI values that have the same age.

    If any NaN values are left over then these can be filled in by setting the `fill_remainder_with_global_func` the
    appropriate function.

    Alternatively you can use the soml.sklearn.SimpleGroupByImputer when you want to do this in a ColumnTransformer.

    Args:
        dataframe: the pd.DatFrame
        column_name: name of the column with the `NaN` values.
        group_by_column_name: name of the column to use as the group by column to calculate the mode
            value of column_name.
        remainder_agg_func: fill any remaining `NaN` values with the global function.
        inplace: update the given dataframe or return a new dataframe.
        add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_groupby_func(dataframe=dataframe, column_name=column_name,
                                         group_by_column_name=group_by_column_name, agg_func='mode',
                                         remainder_agg_func=remainder_agg_func,
                                         inplace=inplace, add_indicator=add_indicator)


def fill_nan_with_groupby_kurt(dataframe: _pd.DataFrame, column_name: str, group_by_column_name: str,
                                  remainder_agg_func='kurt', inplace=True, add_indicator=False) -> _pd.DataFrame | None:
    """
    Fill in missing values based on the kurt value which is calculated on all data having a similar value in
    another column. For example, fill in BMI values based on other BMI values that have the same age.

    If any NaN values are left over then these can be filled in by setting the `fill_remainder_with_global_func` the
    appropriate function.

    Alternatively you can use the soml.sklearn.SimpleGroupByImputer when you want to do this in a ColumnTransformer.

    Args:
        dataframe: the pd.DatFrame
        column_name: name of the column with the `NaN` values.
        group_by_column_name: name of the column to use as the group by column to calculate the kurt
            value of column_name.
        remainder_agg_func: fill any remaining `NaN` values with the global function.
        inplace: update the given dataframe or return a new dataframe.
        add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_groupby_func(dataframe=dataframe, column_name=column_name,
                                         group_by_column_name=group_by_column_name, agg_func='kurt',
                                         remainder_agg_func=remainder_agg_func,
                                         inplace=inplace, add_indicator=add_indicator)


def fill_nan_with_groupby_skew(dataframe: _pd.DataFrame, column_name: str, group_by_column_name: str,
                                  remainder_agg_func='skew', inplace=True, add_indicator=False) -> _pd.DataFrame | None:
    """
    Fill in missing values based on the skew value which is calculated on all data having a similar value in
    another column. For example, fill in BMI values based on other BMI values that have the same age.

    If any NaN values are left over then these can be filled in by setting the `fill_remainder_with_global_func` the
    appropriate function.

    Alternatively you can use the soml.sklearn.SimpleGroupByImputer when you want to do this in a ColumnTransformer.

    Args:
        dataframe: the pd.DatFrame
        column_name: name of the column with the `NaN` values.
        group_by_column_name: name of the column to use as the group by column to calculate the skew
            value of column_name.
        remainder_agg_func: fill any remaining `NaN` values with the global function.
        inplace: update the given dataframe or return a new dataframe.
        add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_groupby_func(dataframe=dataframe, column_name=column_name,
                                         group_by_column_name=group_by_column_name, agg_func='skew',
                                         remainder_agg_func=remainder_agg_func,
                                         inplace=inplace, add_indicator=add_indicator)


def fill_nan_with_groupby_median(dataframe: _pd.DataFrame, column_name: str, group_by_column_name: str,
                                    remainder_agg_func='median', inplace=True, add_indicator=False) -> _pd.DataFrame | None:
    """
    Fill in missing values based on the median value which is calculated on all data having a similar value in
    another column. For example, fill in BMI values based on other BMI values that have the same age.

    If any NaN values are left over then these can be filled in by setting the `fill_remainder_with_global_func` the
    appropriate function.

    Alternatively you can use the soml.sklearn.SimpleGroupByImputer when you want to do this in a ColumnTransformer.

    Args:
        dataframe: the pd.DatFrame
        column_name: name of the column with the `NaN` values.
        group_by_column_name: name of the column to use as the group by column to calculate the median
            value of column_name.
        remainder_agg_func: fill any remaining `NaN` values with the global function.
        inplace: update the given dataframe or return a new dataframe.
        add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_groupby_func(dataframe=dataframe, column_name=column_name,
                                         group_by_column_name=group_by_column_name, agg_func='median',
                                         remainder_agg_func=remainder_agg_func,
                                         inplace=inplace, add_indicator=add_indicator)


def fill_nan_with_groupby_mean(dataframe: _pd.DataFrame, column_name: str, group_by_column_name: str,
                                  remainder_agg_func='mean', inplace=True, add_indicator=False) -> _pd.DataFrame | None:
    """
    Fill in missing values based on the mean value which is calculated on all data having a similar value in
    another column. For example, fill in BMI values based on other BMI values that have the same age.

    If any NaN values are left over then these can be filled in by setting the `fill_remainder_with_global_func` the
    appropriate function.

    Alternatively you can use the soml.sklearn.SimpleGroupByImputer when you want to do this in a ColumnTransformer.

    Args:
        dataframe: the pd.DatFrame
        column_name: name of the column with the `NaN` values.
        group_by_column_name: name of the column to use as the group by column to calculate the mean
            value of column_name.
        remainder_agg_func: fill any remaining `NaN` values with the global function.
        inplace: update the given dataframe or return a new dataframe.
        add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_groupby_func(dataframe=dataframe, column_name=column_name,
                                         group_by_column_name=group_by_column_name, agg_func='mean',
                                         remainder_agg_func=remainder_agg_func,
                                         inplace=inplace, add_indicator=add_indicator)


def fill_nan_with_grouped_max(dataframe: _pd.DataFrame, column_name: str, group_by_column_name: str,
                                 remainder_agg_func='max', inplace=True, add_indicator=False) -> _pd.DataFrame | None:
    """
    Fill in missing values based on the max value which is calculated on all data having a similar value in
    another column. For example, fill in BMI values based on other BMI values that have the same age.

    If any NaN values are left over then these can be filled in by setting the `fill_remainder_with_global_func` the
    appropriate function.

    Alternatively you can use the soml.sklearn.SimpleGroupByImputer when you want to do this in a ColumnTransformer.

    Args:
        dataframe: the pd.DatFrame
        column_name: name of the column with the `NaN` values.
        group_by_column_name: name of the column to use as the group by column to calculate the max
            value of column_name.
        remainder_agg_func: fill any remaining `NaN` values with the global function.
        inplace: update the given dataframe or return a new dataframe.
        add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_groupby_func(dataframe=dataframe, column_name=column_name,
                                         group_by_column_name=group_by_column_name, agg_func='max',
                                         remainder_agg_func=remainder_agg_func,
                                         inplace=inplace, add_indicator=add_indicator)


def fill_nan_with_groupby_min(dataframe: _pd.DataFrame, column_name: str, group_by_column_name: str,
                                 remainder_agg_func='min', inplace=True, add_indicator=False) -> _pd.DataFrame | None:
    """
    Fill in missing values based on the min value which is calculated on all data having a similar value in
    another column. For example, fill in BMI values based on other BMI values that have the same age.

    If any NaN values are left over then these can be filled in by setting the `fill_remainder_with_global_func` the
    appropriate function.

    Alternatively you can use the soml.sklearn.SimpleGroupByImputer when you want to do this in a ColumnTransformer.

    Args:
        dataframe: the pd.DatFrame
        column_name: name of the column with the `NaN` values.
        group_by_column_name: name of the column to use as the group by column to calculate the min
            value of column_name.
        remainder_agg_func: fill any remaining `NaN` values with the global function.
        inplace: update the given dataframe or return a new dataframe.
        add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    return fill_nan_with_groupby_func(dataframe=dataframe, column_name=column_name,
                                         group_by_column_name=group_by_column_name, agg_func='min',
                                         remainder_agg_func=remainder_agg_func,
                                         inplace=inplace, add_indicator=add_indicator)


def fill_nan_with_groupby_func(dataframe: _pd.DataFrame, column_name: str, group_by_column_name: str, agg_func: str,
                                  remainder_agg_func=None, inplace=True, add_indicator=False) -> _pd.DataFrame | None:
    """
    Fill in missing values based on a function  which is calculated on all data having a similar value in
    another column. For example, fill in BMI values based on other BMI values that have the same age.

    If any NaN values are left over then these can be filled in by setting the `fill_remainder_with_global_func` the
    appropriate function.

    Alternatively you can use the soml.sklearn.SimpleGroupByImputer when you want to do this in a ColumnTransformer.

    Args:
        dataframe: the pd.DatFrame
        column_name: name of the column with the `NaN` values.
        group_by_column_name: name of the column to use as the group by column to calculate the function
            value of column_name.
        agg_func: type of function to use: 'mean', 'min', 'max', 'median', 'mode', 'skew' or 'kurt'
        remainder_agg_func: fill any remaining `NaN` values with the global function.
        inplace: update the given dataframe or return a new dataframe.
        add_indicator: add an indicator column indicating with a boolean value if the value was NaN
    """
    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    assert column_name in dataframe, f"fill_nan_with_func_grouped_by: Column '{column_name}' does not exist in dataframe."
    assert group_by_column_name in dataframe, (f"fill_nan_with_func_grouped_by: Column '{group_by_column_name}' "
                                               f"does not exist in dataframe.")

    if add_indicator:
        work_df[column_name + '_nan'] = work_df[column_name].isna().astype(int)

    mean_by_group = work_df.groupby(group_by_column_name)[column_name].agg(agg_func)
    work_df[column_name].fillna(work_df[group_by_column_name].map(mean_by_group), inplace=True)

    # If any other values are still NaN then check if we need to fill these in a default manner.
    if remainder_agg_func:
        c_value = work_df[column_name].apply(remainder_agg_func)
        work_df[column_name].fillna(value=c_value, inplace=True)

    if inplace:
        return None

    return work_df


def generate_code_ordinal_encoder(dataframe: _pd.DataFrame, column_names: list[str]) -> None:
    print(f"\n###################################################\n"
          f"NOTE: The order still needs to be manualy adjusted.\n"
          f"###################################################\n")
    for c in column_names:
        if c in dataframe:
            try:
                values = dataframe[c].unique()
                categories = "',\n    '".join(values)
                print(f"{c}_encoder = sklearn.preprocessing.OrdinalEncoder(categories=[[\n    '{categories}'\n]])")
            except TypeError as e:
                print(f"Column {c} threw an exception {e}.")
        else:
            print(f"generate_code_ordinal_encoder: Column '{c}' does not exist in dataframe.")


def generate_one_hot_encoder(dataframe: _pd.DataFrame, column_names: list[str]) -> None:
    print(f"\n###################################################\n"
          f"NOTE: The order still needs to be manualy adjusted.\n"
          f"###################################################\n")
    for c in column_names:
        if c in dataframe:
            try:
                values = dataframe[c].unique()
                categories = "',\n    '".join(values)
                print(f"{c}_encoder = sklearn.preprocessing.OneHotEncoder(categories=[[\n    '{categories}'\n]])")
            except TypeError as e:
                print(f"Column {c} threw an exception {e}.")
        else:
            print(f"generate_one_hot_encoder: Column '{c}' does not exist in dataframe.")


def delete_rows_not_numeric(dataframe: _pd.DataFrame, column_name: str, inplace=True) -> _pd.DataFrame:
    """
    Deletes all rows containing non-number values inside the given column.

    :param dataframe: the pd.DatFrame
    :param column_name: the column
    :param inplace: return a new instance of the DataFrame (False) or adjust the given DataFrame
    :return: see inplace
    """
    return dataframe.drop(dataframe[_pd.to_numeric(dataframe[column_name], errors='coerce').isna()].index,
                          inplace=inplace)


def drop_columns(dataframe: _pd.DataFrame, column_names: list[str], inplace=True) -> _pd.DataFrame | None:
    """
    Removes (inplace) the given column names from the dataframe.

    :param dataframe: the pd.DatFrame
    :param column_names: the names of the columns to remove
    :param inplace: return a new instance of the DataFrame (False) or adjust the given DataFrame
    """
    if not type(column_names) == list and column_names is not None:
        column_names = [column_names]

    work_df = dataframe
    if not inplace:
        work_df = dataframe.copy(deep=True)

    for c in column_names:
        if c in work_df:
            work_df.drop(c, axis=1, inplace=True)
        else:
            print(f"drop_columns: Column '{c}' does not exist in dataframe.")

    if inplace:
        return None

    return work_df

def describe(dataframe: _pd.DataFrame, column_names: list[str] = None, round=2):
    # If the column_names argument is not a list then create a list
    if not type(column_names) == list and column_names is not None:
        column_names = [column_names]

    # If we don't have a list of column names then create a histogram for every column.
    if column_names is not None:
        columns = list(dataframe[column_names].columns)
    else:
        columns = list(dataframe.columns)

    data = []
    for c in columns:
        v = dataframe[c]

        _mean = _std = _min = _q25 = _q50 = _q75 = _max = z_score_negative = z_score_positive = _irq = _irq_lower = _irq_upper = _np.NAN
        if _pd.api.types.is_numeric_dtype(v):
            _mean = v.mean()
            _std = v.std()
            _min = v.min()
            _q25 = v.quantile(.25)
            _q50 = v.quantile(.50)
            _q75 = v.quantile(.75)
            _max = v.max()

            # Calculate lower value when Z-Score = -3
            z_score_negative = -3 * _std + _mean

            # Calculate upper value when Z-Score = 3
            z_score_positive = 3 * _std + _mean

            # Calculate Interquartile Range (IRQ) and lower / upper bounds
            _irq = _q75 - _q25
            _irq_lower = _q25 - (1.5 * _irq)
            _irq_upper = _q75 + (1.5 * _irq)

        data.append([
            v.name, v.dtype, v.count(), v.isna().sum(), v.nunique(), _mean, _std, z_score_negative, z_score_positive,
            _min, _q25,
            _q50, _q75, _max, _irq_lower, _irq, _irq_upper
        ])

    data = _pd.DataFrame(
        columns=['Column', 'DType', 'NotNull', 'Null', 'Unique', 'Mean', 'Std', '-3σ', '3σ', 'Min', '25%', '50%',
                 '75%', 'Max', 'IRQ-L', 'IRQ', 'IRQ-U'], data=data)

    def highlight_columns(col):
        if col.name == '-3σ' or col.name == 'IRQ-L':
            css = []

            for idx, val in enumerate(col):

                if data['Unique'][idx] > 2 and val > data['Min'][idx]:
                    style = 'color: blue'
                elif _pd.isnull(val):
                    style = 'color: lightgrey'
                else:
                    style = ''

                css.append(style)

            return css;
        elif col.name == '3σ' or col.name == 'IRQ-U':
            css = []

            for idx, val in enumerate(col):
                if data['Unique'][idx] > 3 and val < data['Max'][idx]:
                    style = 'color: blue'
                elif _pd.isnull(val):
                    style = 'color: lightgrey'
                else:
                    style = ''

                css.append(style)

            return css;
        elif col.name == 'Null':
            return ['color: red' if val > 0 else '' for val in col]

        return ['color: lightgrey' if _pd.isnull(val) else '' for val in col]

    return data.style.apply(highlight_columns).format(precision=round)